#include <stdlib.h>
#include <ncurses.h>
#include <locale.h>
#include <getopt.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <string.h>
#include <time.h>

#define KEY_SPACEBAR 32
#define KEY_ESCAPE   27

#ifndef VERSION
#define VERSION "n/a"
#endif

#define PROGRAM_NAME "Termwalk"


struct macase {
  bool up; // has a up connector
  bool left; // has a left connector
  bool down; // has a down connector
  bool right; // has a right connector
  int orient; // orientation can be 0 (up), 1 (left), 2 (down) or 3 (right)
  int is_connected; // connected status: 0 = disconnected, [1-4] = connected to channel [1-4]
  int lastorient; // last printed orientation: prevent refreshing a box if not needed
  int lastcolor; // last color pair: prevent refreshing a box if not needed
  WINDOW *win;
};


// Default grid size
int xmax = 7;
int ymax = 7;

// The keyboard cursor
int cursorx;
int cursory;

// The full grid structure
struct macase board[84][84];




static void sleep_us(unsigned long microseconds) {
  struct timespec ts;
  ts.tv_sec = microseconds / 1000000;
  ts.tv_nsec = (microseconds % 1000000) * 1000;
  nanosleep(&ts, NULL);
}


static bool is_up(struct macase c) {
  if (c.orient == 0) return c.up;
  else if (c.orient == 1) return c.left;
  else if (c.orient == 2) return c.down;
  else if (c.orient == 3) return c.right;
  return 0;
}


static bool is_left(struct macase c) {
  if (c.orient == 0) return c.left;
  else if (c.orient == 1) return c.down;
  else if (c.orient == 2) return c.right;
  else if (c.orient == 3) return c.up;
  return 0;
}


static bool is_down(struct macase c) {
  if (c.orient == 0) return c.down;
  else if (c.orient == 1) return c.right;
  else if (c.orient == 2) return c.up;
  else if (c.orient == 3) return c.left;
  return 0;
}


static bool is_right(struct macase c) {
  if (c.orient == 0) return c.right;
  else if (c.orient == 1) return c.up;
  else if (c.orient == 2) return c.left;
  else if (c.orient == 3) return c.down;
  return 0;
}


static int get_color(int fore, int y, int x) {
  int foreoffset = fore;
  int backoffset = (x+y) % 2;
  if (y == cursory && x == cursorx) backoffset = 2;
  if (fore > 0) {
    foreoffset = (fore + board[ymax/2][xmax/2].orient) % 4 + 1;
  }
  return (COLOR_PAIR(1+(foreoffset*3) + backoffset));
}


static void printbtn(int y, char *str) {
  WINDOW *tmp = newwin(3, 24, y*3 + 1, xmax*5 + 5);
  werase(tmp);
  if (cursorx == xmax && cursory == y) {
    wbkgd(tmp, COLOR_PAIR(6));
  } else {
    wbkgd(tmp, COLOR_PAIR(1));
  }
  box(tmp, 0, 0);
  mvwprintw(tmp, 1, 3, str);
  wrefresh(tmp);
  delwin(tmp);
}


static void draw_menu() {
  printbtn(0, "New game");
  printbtn(1, "Increase grid size");
  printbtn(2, "Decrease grid size");
  printbtn(3, "Solve grid");
  printbtn(4, "Quit game");
}


static void print_one_box(bool solved, int y, int x) {
  if (x > xmax - 1) {
    draw_menu();
    return;
  }
  struct macase c = board[y][x];
  if (solved == false && c.lastorient == c.orient && c.lastcolor == get_color(c.is_connected,y,x)) {
    return;
  }
  bool cup = is_up(c);
  bool cleft = is_left(c);
  bool cdown = is_down(c);
  bool cright = is_right(c);
  WINDOW *tmp = c.win;
  if (solved) {
    cup = c.up;
    cleft = c.left;
    cdown = c.down;
    cright = c.right;
    tmp = newwin(3, 5, y*3 + 1, xmax*5 + 6 + x*5);
  }
  werase(tmp);    
  wbkgd(tmp, get_color(0,y,x));
  wrefresh(tmp);
  
  if (y == ymax/2 && x == xmax/2) {
    mvwprintw(tmp, 1, 1, "\u2590\u2588\u258C");
    if (cup) {
      wattron(tmp, get_color(1,y,x));
      mvwprintw(tmp, 0, 2, "\u25B2");  
      wattroff(tmp, get_color(1,y,x));
    }
    if (cleft) {
      wattron(tmp, get_color(2,y,x));
      mvwprintw(tmp, 1, 0, "\u25C0");  
      wattroff(tmp, get_color(2,y,x));
    }
    if (cdown) {
      wattron(tmp, get_color(3,y,x));
      mvwprintw(tmp, 2, 2, "\u25BC");  
      wattroff(tmp, get_color(3,y,x));
    }
    if (cright) {
      wattron(tmp, get_color(4,y,x));
      mvwprintw(tmp, 1, 4, "\u25B6");  
      wattroff(tmp, get_color(4,y,x));
    }
  }  
  else {
    if (c.is_connected > 0 && !solved) {
      wattron(tmp, get_color(c.is_connected,y,x));
    }
    if (cup) mvwprintw(tmp, 0, 2, "\u2502");
    if (cleft) mvwprintw(tmp, 1, 0, "\u2500\u2500");
    if (cdown) mvwprintw(tmp, 2, 2, "\u2502");
    if (cright) mvwprintw(tmp, 1, 3, "\u2500\u2500");
    if (cup && cleft && cdown && cright) {
      mvwprintw(tmp, 1, 2, "\u253C");
    } else if (cup && cleft && cdown) {
      mvwprintw(tmp, 1, 2, "\u2524");
    } else if (cleft && cdown && cright) {
      mvwprintw(tmp, 1, 2, "\u252C");
    } else if (cdown && cright && cup) {
      mvwprintw(tmp, 1, 2, "\u251C");
    } else if (cright && cup && cleft) {
      mvwprintw(tmp, 1, 2, "\u2534");
    } else if (cup && cleft) {
      mvwprintw(tmp, 1, 2, "\u256F");
    } else if (cup && cdown) {
      mvwprintw(tmp, 1, 2, "\u2502");
    } else if (cup && cright) {
      mvwprintw(tmp, 1, 2, "\u2570");
    } else if (cleft && cdown) {
      mvwprintw(tmp, 1, 2, "\u256E");
    } else if (cleft && cright) {
      mvwprintw(tmp, 1, 2, "\u2500");
    } else if (cdown && cright) {
      mvwprintw(tmp, 1, 2, "\u256D");
    } else {
      wattron(tmp, A_STANDOUT);
      mvwprintw(tmp, 1, 2, " ");
      wattroff(tmp, A_STANDOUT);
    }
    if (c.is_connected > 0 && !solved) {
      wattroff(tmp, get_color(c.is_connected,y,x));
    }
  }

  wrefresh(tmp);
  if (solved) {
    delwin(tmp);
  }
  if (solved == false) {
    if (cursory == y && cursorx == x) {
      board[y][x].lastorient = -1;
      board[y][x].lastcolor = -1;
    } else {
      board[y][x].lastorient = c.orient;
      board[y][x].lastcolor = get_color(c.is_connected,y,x);
    }
  }
}


static bool box_is_empty(int y, int x) {
  if (y < 0 || x < 0 || y > ymax - 1 || x > xmax - 1) {
    return false;
  } 
  if (board[y][x].up == false && board[y][x].left == false && board[y][x].down == false && board[y][x].right == false) {
    return true;
  }
  return false;
}


static bool can_go_up(int y, int x) {
  return (box_is_empty(y-1, x));
}

static bool can_go_left(int y, int x) {
  return (box_is_empty(y, x-1));
}

static bool can_go_down(int y, int x) {
  return (box_is_empty(y+1, x));
}

static bool can_go_right(int y, int x) {
  return (box_is_empty(y, x+1));
}


static bool end_of_the_road(int y, int x) {
  if (can_go_up(y,x) == false && can_go_left(y,x) == false && can_go_down(y,x) == false && can_go_right(y,x) == false) {
    return true;
  }
  return false;
}


static int get_adjacent_non_empty_box(int y, int x) {
  int r = rand() % 4;
  for (int i = 0; i < 2; i++) {
    if (r == 0) {
      if (y > 0 && !box_is_empty(y-1, x)) return 0;
      r++;
    }
    if (r == 1) {
      if (x > 0 && !box_is_empty(y, x-1)) return 1;
      r++;
    }
    if (r == 2) {
      if (y < ymax-1 && !box_is_empty(y+1, x)) return 2;
      r++;
    }
    if (r == 3) {
      if (x < xmax-1 && !box_is_empty(y, x+1)) return 3;
      r = 0;
    }
  }
  return -1;
}


static void draw_board(bool firsttime, bool solved) {
  if (firsttime) {
    WINDOW *tmp;
    if (solved) {
      tmp = newwin(ymax*3 + 2, xmax*5 + 2, 0, xmax*5 + 5);
    } else {
      tmp = newwin(ymax*3 + 2, xmax*5 + 2, 0, 0);
    }
    werase(tmp);
    wbkgd(tmp, COLOR_PAIR(1));
    box(tmp, 0, 0);
    wrefresh(tmp);
    delwin(tmp);
  }
  for (int i = 0; i < ymax; i++) {
    for (int j = 0; j < xmax; j++) {
      if (firsttime) board[i][j].lastorient = -1;
      print_one_box(solved, i, j);
    }
  }
}


static void follow_connected_tree(bool animate, int lastorient, int connected_color, int y, int x) {
  if (y < 0 || x < 0 || y > ymax-1 || x > xmax-1) return;
  if (animate) {
    board[y][x].orient = 0;
  }
  int newcolor = connected_color;
  if (connected_color == 0) newcolor = 1;
  struct macase c = board[y][x];
  if (c.is_connected > 0) return;
  if ((y == ymax/2 && x == xmax/2) || (lastorient == 0 && is_down(c)) || (lastorient == 1 && is_right(c)) || (lastorient == 2 && is_up(c)) || (lastorient == 3 && is_left(c))) {
    board[y][x].is_connected = newcolor;
    if (animate) {
      print_one_box(false, y, x);
      if (ymax>19) {
        sleep_us(3600);
      } else {
        sleep_us((20-ymax) * 3600);
      }
    }    
  } else {
    return;
  }
  if (is_up(c)) {
    if (connected_color == 0) newcolor = 1;
    follow_connected_tree(animate, 0, newcolor, y - 1, x);
  }
  if (is_left(c)) {
    if (connected_color == 0) newcolor = 2;
    follow_connected_tree(animate, 1, newcolor, y, x - 1);
  }
  if (is_down(c)) {
    if (connected_color == 0) newcolor = 3;
    follow_connected_tree(animate, 2, newcolor, y + 1, x);
  }
  if (is_right(c)) {
    if (connected_color == 0) newcolor = 4;
    follow_connected_tree(animate, 3, newcolor, y, x + 1);
  }
}


static void reset_connected_tags(bool animate) {
  for (int i = 0; i < ymax; i++) {
    for (int j = 0; j < xmax; j++) {
      board[i][j].is_connected = 0;
    }
  }
  if (animate) {
    draw_board(false,false);
  }
  follow_connected_tree(animate, 0, 0, ymax/2, xmax/2);  
}


static bool game_won() {
  for (int i = 0; i < ymax; i++) {
    for (int j = 0; j < xmax; j++) {
      struct macase c = board[i][j];
      if (is_up(c) != c.up || is_left(c) != c.left || is_down(c) != c.down || is_right(c) != c.right) {
        return false;
      }
    }
  }
  return true;
}


static void short_help() {
  int tmpmax = ymax;
  if (tmpmax < 5) tmpmax = 5;
  WINDOW *tmp = newwin(3, 76, tmpmax*3 + 2, 0);
  werase(tmp);
  wbkgd(tmp, COLOR_PAIR(1));
  box(tmp, 0, 0);
  mvwprintw(tmp, 1, 2, "Use your mouse or keyboard to connect every server to the central router");
  wrefresh(tmp);
  delwin(tmp);
}


static void version() {
  printf("\n");
  printf("\n");
  printf("Termwalk v%s\n", VERSION);
  printf("Play netwalk game into your terminal\n");
  printf("Copyright 2020 Mobatek (https://mobaxterm.mobatek.net)\n");
  printf("\n");
  printf("\n");
}


static void usage() {
  version();
  printf("\nusage: %s [-v|--version] [-h|--help] [-s ODD_NUMBER|--grid-size=ODD_NUMBER]\n", PROGRAM_NAME);
  printf("\n");
  printf("    -s, --grid-size  Size of the grid (must be an odd number)\n");
  printf("    -v, --version    Show version\n");
  printf("    -h, --help       Show this message\n");
  printf("\n");
  printf("\n");
}




void letsplay(int oldmenu) {
  srand(time(NULL));

  if (xmax > 80 || ymax > 80) {
    xmax = 81;
    ymax = 81;
  }
  if (xmax < 3 || ymax < 3) {
    xmax = 3;
    ymax = 3;
  }
  if (oldmenu < 0) {
    cursory = ymax/2;
    cursorx = xmax/2;
  } else {
    cursory = oldmenu;
    cursorx = xmax;
  }

  int ypath[4];
  int xpath[4];
  for (int i = 0; i < 4; i++) {
    ypath[i] = ymax/2;
    xpath[i] = xmax/2;
  }
  bool start[4] = {false, false, false, false};
  while (!start[0] && !start[1] && !start[2] && !start[3]) {
    for (int i = 0; i < 4; i++) {
      start[i] = (rand() % 2 == 0);
    }
  }

  clear();
  bkgd(COLOR_PAIR(1));
  refresh();

  for (int i = 0; i < ymax; i++) {
    for (int j = 0; j < xmax; j++) {
      board[i][j].up = false;
      board[i][j].left = false;
      board[i][j].down = false;
      board[i][j].right = false;
    }
  }
  
  bool finished = false;
  while (!finished) {
    finished = true;
    for (int i = 0; i < 4; i++) {
      if (start[i] == true && end_of_the_road(ypath[i],xpath[i]) == false) {
        finished = false;
        int r;
        if (ypath[i] == ymax/2 && xpath[i] == xmax/2) {
          r = i;
        } else {
          r = rand() % 4;
        }
        if (r == 0 && board[ypath[i]][xpath[i]].up == false && can_go_up(ypath[i], xpath[i])) {
          board[ypath[i]][xpath[i]].up = true;
          board[ypath[i]-1][xpath[i]].down = true;
          ypath[i]--;
        } else if (r == 1 && board[ypath[i]][xpath[i]].left == false && can_go_left(ypath[i], xpath[i])) {
          board[ypath[i]][xpath[i]].left = true;
          board[ypath[i]][xpath[i]-1].right = true;
          xpath[i]--;
        } else if (r == 2 && board[ypath[i]][xpath[i]].down == false && can_go_down(ypath[i], xpath[i])) {
          board[ypath[i]][xpath[i]].down = true;
          board[ypath[i]+1][xpath[i]].up = true;
          ypath[i]++;
        } else if (r == 3 && board[ypath[i]][xpath[i]].right == false && can_go_right(ypath[i], xpath[i])) {
          board[ypath[i]][xpath[i]].right = true;
          board[ypath[i]][xpath[i]+1].left = true;
          xpath[i]++;
        } else {
          continue;
        }
      }
    }
  }
  
  finished = false;
  while (!finished) {
    finished = true;
    int ypath = -1;
    int xpath = -1;
    for (int i = 0; i < ymax; i++) {
      for (int j = 0; j < xmax; j++) {
        if (box_is_empty(i,j)) {
          int direction = get_adjacent_non_empty_box(i,j);
          if (direction == 0) {
            board[i][j].up = true;
            board[i-1][j].down = true;
          } else if (direction == 1) {
            board[i][j].left = true;
            board[i][j-1].right = true;
          } else if (direction == 2) {
            board[i][j].down = true;
            board[i+1][j].up = true;
          } else if (direction == 3) {
            board[i][j].right = true;
            board[i][j+1].left = true;
          } else {
            continue;
          }
          finished = false;
          ypath = i;
          xpath = j;
        }
      }
    }
    if (finished) break;
    while (!end_of_the_road(ypath,xpath)) {
      int r = rand() % 4;
      if (r == 0 && board[ypath][xpath].up == false && can_go_up(ypath, xpath)) {
        board[ypath][xpath].up = true;
        board[ypath-1][xpath].down = true;
        ypath--;
      } else if (r == 1 && board[ypath][xpath].left == false && can_go_left(ypath, xpath)) {
        board[ypath][xpath].left = true;
        board[ypath][xpath-1].right = true;
        xpath--;
      } else if (r == 2 && board[ypath][xpath].down == false && can_go_down(ypath, xpath)) {
        board[ypath][xpath].down = true;
        board[ypath+1][xpath].up = true;
        ypath++;
      } else if (r == 3 && board[ypath][xpath].right == false && can_go_right(ypath, xpath)) {
        board[ypath][xpath].right = true;
        board[ypath][xpath+1].left = true;
        xpath++;
      } else {
        continue;
      }
    }
  }

  
  for (int i = 0; i < ymax; i++) {
    for (int j = 0; j < xmax; j++) {
      board[i][j].win = newwin(3, 5, i*3 + 1, j*5 + 1);
      board[i][j].lastorient = -1;
      board[i][j].orient = rand() % 4;
    }
  }
  
  reset_connected_tags(false);
  draw_board(true, false);
  draw_menu();
  short_help();
  
  // Main loop
  while (true) {
    int key = getch();
    
    // Mouse buttons
    if (key == KEY_MOUSE) {
      MEVENT event;
      if (getmouse(&event) == OK) {
        if (event.bstate & BUTTON1_CLICKED || event.bstate & BUTTON1_DOUBLE_CLICKED || event.bstate & BUTTON1_PRESSED || event.bstate & BUTTON2_CLICKED || event.bstate & BUTTON2_DOUBLE_CLICKED || event.bstate & BUTTON2_PRESSED) {
          int y = (event.y - 1) / 3;
          int x = (event.x - 1) / 5;
          if (x > xmax - 1) {
            cursorx = xmax;
            cursory = y;
            key = KEY_SPACEBAR;
          } else if (y < ymax && x < xmax) {
            cursory = y;
            cursorx = x;
            key = KEY_SPACEBAR;
            if (event.bstate & BUTTON1_DOUBLE_CLICKED || event.bstate & BUTTON2_DOUBLE_CLICKED) {
              board[y][x].orient = (board[y][x].orient+1) % 4;
            } else if (event.bstate & BUTTON2_CLICKED || event.bstate & BUTTON2_PRESSED) {
              board[y][x].orient = (board[y][x].orient+2) % 4;
            }
          }
        }
      }
    }

    // Action key
    if (key == KEY_SPACEBAR || key == KEY_ENTER || key == 10 || key == '\n') {
      if (cursorx > xmax - 1) {
        if (cursory > 4) cursory = 4;
        if (cursory < 0) cursory = 0;
        if (cursory == 0) key = 'n';        
        if (cursory == 1) key = 'i';        
        if (cursory == 2) key = 'd';        
        if (cursory == 3) key = 's';        
        if (cursory == 4) key = 'q';        
        draw_menu();
        sleep_us(160000);
      } else {
        board[cursory][cursorx].orient = (board[cursory][cursorx].orient+1) % 4;
        reset_connected_tags(false);
        draw_board(false, false);
        draw_menu();
        // If game is won, animate the grid
        if (game_won()) {
          int tmpmax = ymax;
          if (tmpmax < 5) tmpmax = 5;
          WINDOW *tmp = newwin(3, 76, tmpmax*3 + 2, 0);
          werase(tmp);
          wbkgd(tmp, COLOR_PAIR(1));
          wrefresh(tmp);
          delwin(tmp);
          tmp = newwin(3, xmax*5 + 2, ymax*3 + 2, 0);
          werase(tmp);
          wbkgd(tmp, COLOR_PAIR(1));
          box(tmp, 0, 0);
          if (ymax > 5) {
            mvwprintw(tmp, 1, ((xmax * 5 + 2) - 25) / 2, "CONGRATULATIONS, YOU WON!");
          } else {
            mvwprintw(tmp, 1, ((xmax * 5 + 2) - 8) / 2, "YOU WON!");
          }
          wrefresh(tmp);
          delwin(tmp);
          reset_connected_tags(true);  
        }
        continue;
      }
    }

    // Menu keys
    if (key == 'q' || key == 'Q' || key == KEY_ESCAPE) {
      return;
    }
    if (key == 'n' || key == 'N') {
      letsplay(0);
      return;
    }
    if (key == 'i' || key == 'I' || key == '+') {
      xmax = xmax + 2;
      ymax = ymax + 2;
      letsplay(1);
      return;
    }
    if (key == 'd' || key == 'D' || key == '-') {
      xmax = xmax - 2;
      ymax = ymax - 2;
      letsplay(2);
      return;
    }
    if (key == 's' || key == 'S') {
      reset_connected_tags(true);
      continue;
    }
    if (key == KEY_RESIZE) {
      clear();
      refresh();
      draw_board(true, false);
      draw_menu();
      short_help();
      continue;
    }
    
    // Cursor keys
    if (key == KEY_UP || key == KEY_LEFT || key == KEY_DOWN || key == KEY_RIGHT) {
      int prevy = cursory;
      int prevx = cursorx;
      if (key == KEY_UP) cursory--;
      else if (key == KEY_LEFT) cursorx--;
      else if (key == KEY_DOWN) cursory++;
      else if (key == KEY_RIGHT) cursorx++;
      if (cursory < 0) cursory = 0;
      if (cursorx < 0) cursorx = 0;
      if (cursorx > xmax-1) {
        cursorx = xmax;
        if (cursory > 4) cursory = 4;
      } else {
        if (cursory > ymax-1) cursory = ymax-1;
      }
      print_one_box(false, prevy, prevx);
      print_one_box(false, cursory, cursorx);
      continue;
    }
  }
}




int main(int argc, char *argv[]) {
  int option;
  int option_index;
  static const struct option options[] = {
    {"help",       no_argument,       NULL, 'h'},
    {"version",    no_argument,       NULL, 'v'},
    {"grid-size",  required_argument, NULL, 's'},
    {0, 0, 0, 0}
  };

  while ((option = getopt_long(argc, argv, "hvs:", options, &option_index)) != -1) {
    switch (option) {
    case 'v':
      version();
      exit(0);
    case 's':
      xmax = 1 + (atoi(optarg) / 2) * 2;
      ymax = 1 + (atoi(optarg) / 2) * 2;
      break;
    case 'h':
    default:
      usage();
      exit(0);
    }
  }

  // Fixme: this is nasty but we need to put this in order to force ncurses to work
  // in 256 colors mode. For the moment, there is no compatible mode for
  // Termwalk, with 8-colors mode
  putenv("TERM=xterm-256color");

  setlocale(LC_ALL, "");
  initscr();
  raw();
  noecho();
  keypad(stdscr, TRUE);
  start_color();
  curs_set(FALSE);
  set_escdelay(0);
  assume_default_colors(COLOR_WHITE, 234);
  init_pair(1, COLOR_WHITE, 235);
  init_pair(2, COLOR_WHITE, 237);
  init_pair(3, COLOR_WHITE, 232);
  init_pair(4, COLOR_BLUE, 235);
  init_pair(5, COLOR_BLUE, 237);
  init_pair(6, COLOR_BLUE, 232);
  init_pair(7, COLOR_GREEN, 235);
  init_pair(8, COLOR_GREEN, 237);
  init_pair(9, COLOR_GREEN, 232);
  init_pair(10, COLOR_YELLOW, 235);
  init_pair(11, COLOR_YELLOW, 237);
  init_pair(12, COLOR_YELLOW, 232);
  init_pair(13, COLOR_MAGENTA, 235);
  init_pair(14, COLOR_MAGENTA, 237);
  init_pair(15, COLOR_MAGENTA, 232);
  mousemask(ALL_MOUSE_EVENTS, NULL);

  letsplay(-1);

  endwin();

  return 0;
}
